// -*- C++ -*-
/*!
 * @file InPort.h
 * @brief InPort template class
 * @date $Date: 2007/07/20 15:54:50 $
 * @author Noriaki Ando <n-ando@aist.go.jp>
 *
 * Copyright (C) 2003-2005
 *     Task-intelligence Research Group,
 *     Intelligent Systems Research Institute,
 *     National Institute of
 *         Advanced Industrial Science and Technology (AIST), Japan
 *     All rights reserved.
 *
 * $Id: InPort.h,v 1.6.4.1 2007/07/20 15:54:50 n-ando Exp $
 *
 */

/*
 * $Log: InPort.h,v $
 * Revision 1.6.4.1  2007/07/20 15:54:50  n-ando
 * Now ACE_OS::gettimeofday() is used for win32 porting.
 *
 * Revision 1.6  2007/01/06 17:47:51  n-ando
 * Some changes.
 * - Callback declarations are changed.
 * - Return value of write().
 *
 * Revision 1.5  2006/12/02 18:37:29  n-ando
 * A trivial fix.
 *
 * Revision 1.4  2006/11/07 09:26:12  n-ando
 * Renamed RtcInPort.h to InPort.h.
 *
 * Revision 1.3  2005/05/27 07:29:32  n-ando
 * - InPort/OutPort interface was changed.
 *   Exception RTM::InPort::Disconnected -> RTM:PortBase::Disconnected
 *
 * Revision 1.2  2005/05/16 06:12:15  n-ando
 * - Time variables of "m_Value" were initialized in constructors.
 *
 * Revision 1.1.1.1  2005/05/12 09:06:18  n-ando
 * Public release.
 *
 *
 */

#ifndef RtcInPort_h
#define RtcInPort_h

#include <string>
#include <vector>
#include <iostream>
//#include <rtm/InPortBase.h>
#include <rtm/BufferBase.h>
#include <rtm/PortCallBack.h>
#include <rtm/RTC.h>

#define TIMEOUT_TICK_USEC 10
#define USEC_PER_SEC 1000000

namespace RTC
{
  /*!
   * @if jp
   *
   * @class InPort
   *
   * @brief InPort ev[gNX
   * 
   * InPort ̎ł InPortAny<T> ̃ev[gNXB
   * <T> RTCDataType.idl ɂĒ`Ă^ŁAoƂ
   * Time ^ tm ,  T^ data \̂łȂĂ͂ȂȂB
   * InPort ͓ɃOobt@AO瑗Mꂽf[^
   * ̃Oobt@Ɋi[BOobt@̃TCY̓ftHg64
   * ȂĂ邪ARXgN^ɂTCYw肷邱ƂłB
   * f[^̓tOɂĖǁAǏԂǗAisNew(), getNewDataLen()
   * getNewList(), getNewListReverse() ̃\bhɂnhO邱Ƃ
   * łB
   *
   * @else
   *
   * @class InPortAny
   *
   * @brief OutPort template class
   *
   * This class template provides interfaces to input port.
   * Component developer can define input value, which act as input
   * port from other components, using this template.
   * This is class template. This class have to be incarnated class as port
   * value types. This value types are previously define RtComponent IDL.
   * ex. type T: TimedFload, TimedLong etc... 
   *
   * @endif
   */
  template <class DataType,
	    template <class DataType> class Buffer = NullBuffer >
  class InPort
    : public Buffer<DataType>
  {
  public:
    /*!
     * @if jp
     *
     * @brief InPortAny NXRXgN^
     *
     * InPortAny<T> NX̃RXgN^B
     * p[^Ƃė^ T ^̕ϐɃoChB
     *
     * @param name InPort BInPortBase:name() ɂQƂB
     * @param value  InPort ɃoCh T ^̕ϐ
     * @param bufsize InPort ̃Oobt@̃obt@
     *
     * @else
     *
     * @brief A constructor.
     *
     * Setting channel name and registering channel value.
     *
     * @param name A name of the InPort. This name is referred by
     *             InPortBase::name().
     * @param value A channel value related with the channel.
     * @param bufsize Buffer length of internal ring buffer of InPort 
     *
     * @endif
     */
    InPort(const char* name, DataType& value,
	   int bufsize=64, 
	   bool read_block = false, bool write_block = false,
	   int read_timeout = 0, int write_timeout = 0)
      : Buffer<DataType>(bufsize),
	m_name(name), m_value(value),
	m_readBlock(read_block),   m_readTimeout(read_timeout),
	m_writeBlock(write_block), m_writeTimeout(write_timeout),
	m_OnWrite(NULL), m_OnWriteConvert(NULL),
	m_OnRead(NULL),  m_OnReadConvert(NULL),
	m_OnOverflow(NULL), m_OnUnderflow(NULL)
    {
    };
    
    
    /*!
     * @if jp
     *
     * @brief InPortAny NXfXgN^
     *
     * InPortAny<T> NX̃fXgN^B
     *
     * @else
     *
     * @brief A destructor
     *
     * @endif
     */
    virtual ~InPort(){};



    virtual const char* name()
    {
      return m_name.c_str();
    }


    /*!
     * @if jp
     *
     * @brief DataPort ɒl
     *
     * DataPort ɒlށB
     *
     * - R[obNt@N^ OnWrite ZbgĂꍇA
     *   InPort ێobt@ɏޑO OnWrite Ă΂B
     * - InPort ێobt@I[o[t[ołobt@łA
     *   AލۂɃobt@I[o[t[oꍇA
     *   R[obNt@N^ OnOverflow Ă΂B
     * - R[obNt@N^ OnWriteConvert ZbgĂꍇA
     *   obt@ݎɁAOnWriteConvert  operator()() ̖߂l
     *   obt@ɏ܂B
     * - setWriteTimeout() ɂ菑ݎ̃^CAEgݒ肳ĂꍇA
     *   ^CAEgԂobt@tԂ̂҂A
     *   OnOverflowZbgĂ΂ĂяoĖ߂B
     *
     * @else
     *
     * @brief 
     *
     * @endif
     */
    bool write(const DataType& value)
    {
      if (m_OnWrite != NULL) (*m_OnWrite)(value);      

      long int timeout = m_writeTimeout;

      timeval tm_cur, tm_pre;
      ACE_Time_Value tt;
      tt = ACE_OS::gettimeofday();
      tm_pre = tt.operator timeval();

      // blocking and timeout wait
      while (m_writeBlock && this->isFull())
	{
	  if (m_writeTimeout < 0) 
	    {
	      usleep(TIMEOUT_TICK_USEC);
	      continue;
	    }

	  // timeout wait
	  ACE_Time_Value tt;
	  tt = ACE_OS::gettimeofday();
	  tm_cur = tt.operator timeval();
	  long int sec (tm_cur.tv_sec  - tm_pre.tv_sec);
	  long int usec(tm_cur.tv_usec - tm_pre.tv_usec);

	  timeout -= (sec * USEC_PER_SEC + usec);
	  if (timeout < 0) break;

	  tm_pre = tm_cur;
	  usleep(TIMEOUT_TICK_USEC);
	}

      if (this->isFull() && m_OnOverflow != NULL)
	{
	  (*m_OnOverflow)(value);
	  return false;
	}

      if (m_OnWriteConvert == NULL) 
	{
	  this->put(value);
	}
      else
	{
	  this->put((*m_OnWriteConvert)(value));
	}
      return true;
    }
    

    /*!
     * @if jp
     *
     * @brief DataPort lǂݏo
     *
     * DataPort lǂݏo
     *
     * - R[obNt@N^ OnRead ZbgĂꍇA
     *   DataPort ێobt@ǂݏoO OnRead Ă΂B
     * - DataPort ێobt@A_[t[ołobt@ŁA
     *   AǂݏoۂɃobt@A_[t[oꍇA
     *   R[obNt@N^ OnUnderflow Ă΂B
     * - R[obNt@N^ OnReadConvert ZbgĂꍇA
     *   obt@ݎɁAOnReadConvert  operator()() ̖߂l
     *   read()̖߂lƂȂB
     * - setReadTimeout() ɂǂݏõ^CAEgݒ肳ĂꍇA
     *   obt@A_[t[Ԃ܂Ń^CAEgԂ҂A
     *   OnUnderflowZbgĂ΂ĂяoĖ߂
     *
     * @else
     *
     * @brief [CORBA interface] Put data on InPort
     *
     * @endif
     */
    DataType read()
    {
      if (m_OnRead != NULL) (*m_OnRead)();      

      long int timeout = m_readTimeout;

      timeval tm_cur, tm_pre;
      ACE_Time_Value tt;
      tt = ACE_OS::gettimeofday();
      tm_pre = tt.operator timeval();

      // blocking and timeout wait
      while (m_readBlock && this->isEmpty())
	{
	  if (m_readTimeout < 0)
	    {
	      usleep(TIMEOUT_TICK_USEC);
	      continue;
	    }

	  // timeout wait
	  ACE_Time_Value tt;
	  tt = ACE_OS::gettimeofday();
	  tm_cur = tt.operator timeval();
	  long int sec (tm_cur.tv_sec  - tm_pre.tv_sec);
	  long int usec(tm_cur.tv_usec - tm_pre.tv_usec);

	  timeout -= (sec * USEC_PER_SEC + usec);
	  if (timeout < 0) break;

	  tm_pre = tm_cur;
	  usleep(TIMEOUT_TICK_USEC);
	}

      if (this->isEmpty() && m_OnUnderflow != NULL)
	{
	  m_value = (*m_OnUnderflow)();
	  return m_value;
	}

      if (m_OnReadConvert == NULL) 
	{
	  m_value = this->get();
	  return m_value;
	}
      else
	{
	  m_value = (*m_OnReadConvert)(this->get());
	  return m_value;
	}
      // never comes here
      return m_value;
    }



    /*!
     * @if jp
     *
     * @brief InPort ̃Oobt@̒l
     *
     * InPort ̃Oobt@̒lB
     *
     * @else
     *
     * @brief Initialize ring buffer value of InPort
     *
     * @endif
     */
    virtual void init(DataType& value)
    {
      //      m_buffer.init(value);
    }
    
    
    /*!
     * @if jp
     *
     * @brief oChꂽ T ^̕ϐ InPort obt@̍ŐVlǂݍ
     *
     * oChꂽ T ^̃f[^ InPort ̍ŐVlǂݍށB
     * RXgN^ T ^̕ϐ InPort oChĂȂ΂ȂȂB
     * ̃\bh̓|[tBbNɎgp鎖OƂĂ邽߁A
     * ^ɈˑȂA߂lƂȂĂB
     *
     * @else
     *
     * @brief Read into bound T-type data from current InPort
     *
     * @endif
     */
    void update()
    {
      try
	{
	  m_value = this->get();
	}
      catch (...)
	{
	  if (m_OnUnderflow != NULL) (*m_OnUnderflow)();
	}
      return;
    };
    
    
    /*!
     * @if jp
     *
     * @brief T ^̃f[^ InPort ̍ŐVlf[^ǂݍ
     *
     * @param rhs InPort obt@lǂݍ T ^ϐ
     *
     * @else
     *
     * @brief Read from InPort buffer to type-T variable
     *
     * @param rhs type-T variable which is wrote from InPort buffer
     *
     * @endif
     */
    void operator>>(DataType& rhs)
    {
      rhs = read();
      return;
    }
    

    void operator<<(DataType& value)
    {
      write(value);
      return;
    }

    
    /*!
     * @if jp
     *
     * @brief ǂ̐Vf[^擾
     *
     * @else
     *
     * @brief Get number of new data to be read.
     *
     * @endif
     */
    /*
    virtual int getNewDataLen()
    {
      return m_buffer->new_data_len();
    }
    */    
    
    /*!
     * @if jp
     *
     * @brief ǂ̐Vf[^擾
     *
     * @else
     *
     * @brief Get new data to be read.
     *
     * @endif
     */
    /*
    virtual std::vector<T> getNewList()
    {
      return m_buffer.get_new_list();
    }
    */
    
    /*!
     * @if jp
     *
     * @brief ǂ̐Vf[^t(V->)Ŏ擾
     *
     * @else
     *
     * @brief Get new data to be read.
     *
     * @endif
     */
    /*
    virtual std::vector<T> getNewListReverse()
    {
      return m_buffer.get_new_rlist();
    }
    */

    /*!
     * @if jp
     *
     * @brief InPort obt@Ƀf[^͎̃R[obN̐ݒ
     *
     * InPort obt@Ƀf[^putꂽƂɌĂ΂R[obN
     * IuWFNgݒ肷Bݒ肳R[obNIuWFNg
     * InPort<DataType>::OnPutNXpA const DataType& A
     * ߂l void  operator() ֐ĂKvB
     *
     * struct MyOnPutCallback : public InPort<DataType> {<br>
     *   void operator()(const DataType data) {<br>
     *     <br>
     *   }<br>
     * };<br>
     * ̂悤ɃR[obNIuWFNgA<br>
     * <br> 
     * m_inport.setOnPut(new MyOnPutCallback());<br>
     * ̂悤ɃR[obNIuWFNgZbgB
     *
     * @else
     *
     * @brief Get new data to be read.
     *
     * @endif
     */
    inline void setOnWrite(OnWrite<DataType>* on_write)
    {
      m_OnWrite = on_write;
    }

    inline void setOnWriteConvert(OnWriteConvert<DataType>* on_wconvert)
    {
      m_OnWriteConvert = on_wconvert;
    }

    inline void setOnRead(OnRead<DataType>* on_read)
    {
      m_OnRead = on_read;
    }

    inline void setOnReadConvert(OnReadConvert<DataType>* on_rconvert)
    {
      m_OnReadConvert = on_rconvert;
    }

    inline void setOnOverflow(OnOverflow<DataType>* on_overflow)
    {
      m_OnOverflow = on_overflow;
    }

    inline void setOnUnderflow(OnUnderflow<DataType>* on_underflow)
    {
      m_OnUnderflow = on_underflow;
    }


  private:
    /*!
     * @if jp
     * @brief |[g
     * @else
     * @berif Port's name
     * @else
     */
    std::string m_name;

    /*!
     * @if jp
     * @brief oCh T ^̕ϐւ̎Q
     * @else
     * @brief reference to type-T value bound this OutPort
     * @endif
     */
    DataType& m_value;
    
    /*!
     * @if jp
     * @brief obt@
     * @else
     * @brief Buffer
     * @endif
     */
    //    Buffer<DataType> m_buffer;

    bool m_readBlock;
    long int m_readTimeout;
    bool m_writeBlock;
    long int m_writeTimeout;

    /*!
     * @if jp
     * @brief OnWrite R[obNt@N^ւ̃|C^
     * @else
     * @brief OnWrite callback functor pointer
     * @endif
     */
    OnWrite<DataType>* m_OnWrite;

    /*!
     * @if jp
     * @brief OnWriteConvert R[obNt@N^ւ̃|C^
     * @else
     * @brief OnWriteConvert callback functor pointer
     * @endif
     */
    OnWriteConvert<DataType>* m_OnWriteConvert;

    /*!
     * @if jp
     * @brief OnRead R[obNt@N^ւ̃|C^
     * @else
     * @brief OnRead callback functor pointer
     * @endif
     */
    OnRead<DataType>* m_OnRead;

    /*!
     * @if jp
     * @brief OnReadConvert R[obNt@N^ւ̃|C^
     * @else
     * @brief OnReadConvert callback functor pointer
     * @endif
     */
    OnReadConvert<DataType>* m_OnReadConvert;

    /*!
     * @if jp
     * @brief OnOverflow R[obNt@N^ւ̃|C^
     * @else
     * @brief OnOverflow callback functor pointer
     * @endif
     */
    OnOverflow<DataType>* m_OnOverflow;

    /*!
     * @if jp
     * @brief OnUnderflow R[obNt@N^ւ̃|C^
     * @else
     * @brief OnUnderflow callback functor pointer
     *
     * @endif
     */
    OnUnderflow<DataType>* m_OnUnderflow;

    
  };
  
}; // End of namesepace RTM

#endif // RtcInPort_h
